==> infima_client-0.186.0-py3-none-any.whl: infima_client/search_builder.py <==
"""SearchBuilder creates queries to be used in the pool search."""

from typing import List, Union

STATES = [
    "AL",
    "AK",
    "AZ",
    "AR",
    "CA",
    "CO",
    "CT",
    "DC",
    "DE",
    "FL",
    "GA",
    "HI",
    "ID",
    "IL",
    "IN",
    "IA",
    "KS",
    "KY",
    "LA",
    "ME",
    "MD",
    "MA",
    "MI",
    "MN",
    "MS",
    "MO",
    "MT",
    "NE",
    "NV",
    "NH",
    "NJ",
    "NM",
    "NY",
    "NC",
    "ND",
    "OH",
    "OK",
    "OR",
    "PA",
    "RI",
    "SC",
    "SD",
    "TN",
    "TX",
    "UT",
    "VT",
    "VA",
    "WA",
    "WV",
    "WI",
    "WY",
]


class AttributeRange:
    def __init__(self, name: str, min: float, max: float):
        self.name = name
        self.min = min
        self.max = max

    def to_dict(self):
        out = {}
        if self.min:
            out["min"] = self.min
        if self.max:
            out["max"] = self.max
        return out


class AttributeDateRange:
    def __init__(self, name: str, start: int, end: int):
        self.name = name
        self.start = start
        self.end = end

    def to_dict(self):
        out = {}
        if self.start and isinstance(self.start, int):
            out["start"] = {"year": self.start, "month": 1, "day": 1}
        if self.end and isinstance(self.start, int):
            out["end"] = {"year": self.end, "month": 12, "day": 31}
        return out


ATTRIBUTES = {
    "coupon": lambda min, max: AttributeRange("coupon_range", min, max),
    "upb": lambda min, max: AttributeRange("upb_range", min, max),
    "upb issuance": lambda min, max: AttributeRange("iss_upb_range", min, max),
    "waocs": lambda min, max: AttributeRange("waocs_range", min, max),
    "waoltv": lambda min, max: AttributeRange("waoltv_range", min, max),
    "waols": lambda min, max: AttributeRange("waols_range", min, max),
    "minimum original loan balance": lambda min, max: AttributeRange(
        "min_iss_loan_balance_range", min, max
    ),
    "maximum original loan balance": lambda min, max: AttributeRange(
        "max_iss_loan_balance_range", min, max
    ),
    "minimum credit score": lambda min, max: AttributeRange("min_cs_range", min, max),
    "maximum credit score": lambda min, max: AttributeRange("max_cs_range", min, max),
    "wac": lambda min, max: AttributeRange("wac_range", min, max),
    "issue date": lambda min, max: AttributeDateRange("issue_date_range", min, max),
    "maturity date": lambda min, max: AttributeDateRange(
        "maturity_date_range", min, max
    ),
    "investor percentage": lambda min, max: AttributeRange(
        "investor_occ_range", min, max
    ),
}


class InputRange:
    def __init__(self, name: str, min_proportion: float, max_proportion: float):
        self.name = name
        self.min_proportion = min_proportion
        self.max_proportion = max_proportion

    def to_dict(self):
        return {
            "name": self.name,
            "min_proportion": self.min_proportion,
            "max_proportion": self.max_proportion,
        }


def validate_proportion_range(rg: str, min_proportion: float, max_proportion: float):
    if not 0 <= min_proportion <= 100:
        raise ValueError(f"{rg} min_proportion should be between 0 and 100")
    if not 0 <= max_proportion <= 100:
        raise ValueError(f"{rg} max_proportion should be between 0 and 100")
    if not min_proportion < max_proportion:
        raise ValueError(f"{rg} min_proportion should be less than max_proportion")


def validate_range(rg: str, min: float, max: float):
    if min and max and not min <= max:
        raise ValueError(f"{rg} min should be less than max")


class SearchBuilder:
    def __init__(self, cohorts: List[str] = None):
        """


        Args:
            cohorts: List of cohorts providing the universe in which to search for pools.

        Returns:
            A query that can be manipulated to add search criteria.
        """
        self.cohorts = cohorts
        self.included_servicers = []
        self.excluded_servicers = []
        self.included_geos = []
        self.excluded_geos = []
        self.attributes = {}

    def with_criteria(self):
        return (
            self.with_servicers_criteria()
            or self.with_geos_criteria()
            or self.with_attributes_criteria()
        )

    def with_servicers_criteria(self) -> bool:
        return len(self.included_servicers) + len(self.excluded_servicers) > 0

    def with_geos_criteria(self) -> bool:
        return len(self.included_geos) + len(self.excluded_geos) > 0

    def with_attributes_criteria(self) -> bool:
        return len(self.attributes) > 0

    def with_servicers(
        self, servicer: str, min_proportion: float = 0, max_proportion: float = 100
    ):

        """
        Args:
            servicer: Servicers on which to select with proportions required.
            min_proportion: Minimum proportion required for that servicer, default to 0.
            max_proportion: Maximum proportion required for that servicer, default to 100.

        Returns:
            The query itself so these calls and other can be chained.
        """
        validate_proportion_range(
            f"including servicer: {servicer}", min_proportion, max_proportion
        )

        self.included_servicers.append(
            InputRange(
                name=servicer,
                min_proportion=min_proportion,
                max_proportion=max_proportion,
            )
        )
        return self

    def excluding_servicers(
        self, servicer: str, min_proportion: float = 0, max_proportion: float = 100
    ):
        """
        Args:
            servicer: Servicers on which to exclude with proportions required.
            min_proportion: Minimum proportion required for that servicer, default to 0.
            max_proportion: Maximum proportion required for that servicer, default to 100.

        Returns:
            The query itself so these calls and other can be chained.
        """
        validate_proportion_range(
            f"excluding servicer: {servicer}", min_proportion, max_proportion
        )

        self.excluded_servicers.append(
            InputRange(
                name=servicer,
                min_proportion=min_proportion,
                max_proportion=max_proportion,
            )
        )
        return self

    def with_geos(
        self, state: str, min_proportion: float = 0, max_proportion: float = 100
    ):
        """
        Args:
            state: US State (abbreviation) on which to select with proportions required.
            min_proportion: Minimum proportion required for that state, default to 0.
            max_proportion: Maximum proportion required for that state, default to 100.

        Returns:
            The query itself so these calls and other can be chained.
        """
        if state not in STATES:
            raise ValueError(
                f"{state} is not a valid state, use abbreviation like CA, ..."
            )
        validate_proportion_range(
            f"including state: {state}", min_proportion, max_proportion
        )

        self.included_geos.append(
            InputRange(
                name=state,
                min_proportion=min_proportion,
                max_proportion=max_proportion,
            )
        )
        return self

    def excluding_geos(
        self, state: str, min_proportion: float = 0, max_proportion: float = 100
    ):
        """
        Args:
            state: US State (abbreviation) on which to exclude with proportions required.
            min_proportion: Minimum proportion required for that state, default to 0.
            max_proportion: Maximum proportion required for that state, default to 100.

        Returns:
            The query itself so these calls and other can be chained.
        """
        if state not in STATES:
            raise ValueError(
                f"{state} is not a valid state, use abbreviation like CA, ..."
            )
        validate_proportion_range(
            f"excluding state: {state}", min_proportion, max_proportion
        )

        self.excluded_geos.append(
            InputRange(
                name=state,
                min_proportion=min_proportion,
                max_proportion=max_proportion,
            )
        )
        return self

    def available_attributes(self):
        """
        Returns:
            List of all available attributes on which to filter.
        """
        return list(ATTRIBUTES.keys())

    def with_attributes(
        self,
        attribute: str,
        min: Union[float, int] = None,
        max: Union[float, int] = None,
    ):
        """
        Args:
            state: Pool attribute to filter on. For all possible attributes, run query.available_attributes()
            min_proportion: Minimum required for that attribute.
            max_proportion: Maximum required for that attribute.

        Returns:
            The query itself so these calls and other can be chained.
        """
        validate_range(f"attribute: {attribute}", min, max)
        if attribute not in ATTRIBUTES:
            raise ValueError(
                f"unknown attributes {attribute}, available are {','.join(ATTRIBUTES.keys())}"
            )
        self.attributes[attribute] = ATTRIBUTES[attribute](min, max)
        return self

    def to_dict(self):
        out = {}
        if self.cohorts is not None:
            out["cohorts"] = {"cohorts": self.cohorts}
        if self.with_criteria():
            out["criteria"] = {}
        if self.with_servicers_criteria():
            servicers = {}
            if self.included_servicers:
                servicers["includes"] = [
                    inp.to_dict() for inp in self.included_servicers
                ]
            if self.excluded_servicers:
                servicers["excludes"] = [
                    inp.to_dict() for inp in self.excluded_servicers
                ]
            out["criteria"]["servicer"] = servicers
        if self.with_geos_criteria():
            geos = {}
            if self.included_geos:
                geos["includes"] = [inp.to_dict() for inp in self.included_geos]
            if self.excluded_geos:
                geos["excludes"] = [inp.to_dict() for inp in self.excluded_geos]
            out["criteria"]["geo"] = geos
        if self.with_attributes_criteria():
            out["attributes"] = {
                rg.name: rg.to_dict() for attribute, rg in self.attributes.items()
            }
        return out

==> infima_client-0.259.0-py3-none-any.whl: infima_client/search_builder.py <==
"""SearchBuilder creates queries to be used in the pool search."""

from typing import List, Union

STATES = [
    "AL",
    "AK",
    "AZ",
    "AR",
    "CA",
    "CO",
    "CT",
    "DC",
    "DE",
    "FL",
    "GA",
    "HI",
    "ID",
    "IL",
    "IN",
    "IA",
    "KS",
    "KY",
    "LA",
    "ME",
    "MD",
    "MA",
    "MI",
    "MN",
    "MS",
    "MO",
    "MT",
    "NE",
    "NV",
    "NH",
    "NJ",
    "NM",
    "NY",
    "NC",
    "ND",
    "OH",
    "OK",
    "OR",
    "PA",
    "RI",
    "SC",
    "SD",
    "TN",
    "TX",
    "UT",
    "VT",
    "VA",
    "WA",
    "WV",
    "WI",
    "WY",
]


class AttributeRange:
    def __init__(self, name: str, min: float, max: float):
        self.name = name
        self.min = min
        self.max = max

    def to_dict(self):
        out = {}
        if self.min:
            out["min"] = self.min
        if self.max:
            out["max"] = self.max
        return out


class AttributeDateRange:
    def __init__(self, name: str, start: int, end: int):
        self.name = name
        self.start = start
        self.end = end

    def to_dict(self):
        out = {}
        if self.start and isinstance(self.start, int):
            out["start"] = {"year": self.start, "month": 1, "day": 1}
        if self.end and isinstance(self.start, int):
            out["end"] = {"year": self.end, "month": 12, "day": 31}
        return out


ATTRIBUTES = {
    "coupon": lambda min, max: AttributeRange("coupon_range", min, max),
    "upb": lambda min, max: AttributeRange("upb_range", min, max),
    "upb issuance": lambda min, max: AttributeRange("iss_upb_range", min, max),
    "waocs": lambda min, max: AttributeRange("waocs_range", min, max),
    "waoltv": lambda min, max: AttributeRange("waoltv_range", min, max),
    "waols": lambda min, max: AttributeRange("waols_range", min, max),
    "minimum original loan balance": lambda min, max: AttributeRange(
        "min_iss_loan_balance_range", min, max
    ),
    "maximum original loan balance": lambda min, max: AttributeRange(
        "max_iss_loan_balance_range", min, max
    ),
    "minimum credit score": lambda min, max: AttributeRange("min_cs_range", min, max),
    "maximum credit score": lambda min, max: AttributeRange("max_cs_range", min, max),
    "wac": lambda min, max: AttributeRange("wac_range", min, max),
    "cpr": lambda min, max: AttributeRange("cpr_range", min, max),
    "investor percentage": lambda min, max: AttributeRange(
        "investor_occ_range", min, max
    ),
    "issue date": lambda min, max: AttributeDateRange("issue_date_range", min, max),
    "maturity date": lambda min, max: AttributeDateRange(
        "maturity_date_range", min, max
    ),
}

PREDICTIONS = {
    "1m predicted cpr": lambda min, max: AttributeRange("pred_1m_cpr_range", min, max),
    "1m predicted cpr delta": lambda min, max: AttributeRange(
        "pred_1m_cpr_delta_range", min, max
    ),
    "3m predicted cpr": lambda min, max: AttributeRange("pred_3m_cpr_range", min, max),
    "6m predicted cpr": lambda min, max: AttributeRange("pred_6m_cpr_range", min, max),
}


class InputRange:
    def __init__(self, name: str, min_proportion: float, max_proportion: float):
        self.name = name
        self.min_proportion = min_proportion
        self.max_proportion = max_proportion

    def to_dict(self):
        return {
            "name": self.name,
            "min_proportion": self.min_proportion,
            "max_proportion": self.max_proportion,
        }


def validate_proportion_range(rg: str, min_proportion: float, max_proportion: float):
    if not 0 <= min_proportion <= 100:
        raise ValueError(f"{rg} min_proportion should be between 0 and 100")
    if not 0 <= max_proportion <= 100:
        raise ValueError(f"{rg} max_proportion should be between 0 and 100")
    if not min_proportion < max_proportion:
        raise ValueError(f"{rg} min_proportion should be less than max_proportion")


def validate_range(rg: str, min: float, max: float):
    if min and max and not min <= max:
        raise ValueError(f"{rg} min should be less than max")


class SearchBuilder:
    def __init__(self, cohorts: List[str] = None):
        """


        Args:
            cohorts: List of cohorts providing the universe in which to search for pools.

        Returns:
            A query that can be manipulated to add search criteria.
        """
        self.cohorts = cohorts
        self.included_servicers = []
        self.excluded_servicers = []
        self.included_geos = []
        self.excluded_geos = []
        self.attributes = {}
        self.predictions = {}

    def with_criteria(self):
        return (
            self.with_servicers_criteria()
            or self.with_geos_criteria()
            or self.with_attributes_criteria()
            or self.with_predictions_criteria()
        )

    def with_servicers_criteria(self) -> bool:
        return len(self.included_servicers) + len(self.excluded_servicers) > 0

    def with_geos_criteria(self) -> bool:
        return len(self.included_geos) + len(self.excluded_geos) > 0

    def with_attributes_criteria(self) -> bool:
        return len(self.attributes) > 0

    def with_predictions_criteria(self) -> bool:
        return len(self.predictions) > 0

    def with_servicers(
        self, servicer: str, min_proportion: float = 0, max_proportion: float = 100
    ):

        """
        Args:
            servicer: Servicers on which to select with proportions required.
            min_proportion: Minimum proportion required for that servicer, default to 0.
            max_proportion: Maximum proportion required for that servicer, default to 100.

        Returns:
            The query itself so these calls and other can be chained.
        """
        validate_proportion_range(
            f"including servicer: {servicer}", min_proportion, max_proportion
        )

        self.included_servicers.append(
            InputRange(
                name=servicer,
                min_proportion=min_proportion,
                max_proportion=max_proportion,
            )
        )
        return self

    def excluding_servicers(
        self, servicer: str, min_proportion: float = 0, max_proportion: float = 100
    ):
        """
        Args:
            servicer: Servicers on which to exclude with proportions required.
            min_proportion: Minimum proportion required for that servicer, default to 0.
            max_proportion: Maximum proportion required for that servicer, default to 100.

        Returns:
            The query itself so these calls and other can be chained.
        """
        validate_proportion_range(
            f"excluding servicer: {servicer}", min_proportion, max_proportion
        )

        self.excluded_servicers.append(
            InputRange(
                name=servicer,
                min_proportion=min_proportion,
                max_proportion=max_proportion,
            )
        )
        return self

    def with_geos(
        self, state: str, min_proportion: float = 0, max_proportion: float = 100
    ):
        """
        Args:
            state: US State (abbreviation) on which to select with proportions required.
            min_proportion: Minimum proportion required for that state, default to 0.
            max_proportion: Maximum proportion required for that state, default to 100.

        Returns:
            The query itself so these calls and other can be chained.
        """
        if state not in STATES:
            raise ValueError(
                f"{state} is not a valid state, use abbreviation like CA, ..."
            )
        validate_proportion_range(
            f"including state: {state}", min_proportion, max_proportion
        )

        self.included_geos.append(
            InputRange(
                name=state,
                min_proportion=min_proportion,
                max_proportion=max_proportion,
            )
        )
        return self

    def excluding_geos(
        self, state: str, min_proportion: float = 0, max_proportion: float = 100
    ):
        """
        Args:
            state: US State (abbreviation) on which to exclude with proportions required.
            min_proportion: Minimum proportion required for that state, default to 0.
            max_proportion: Maximum proportion required for that state, default to 100.

        Returns:
            The query itself so these calls and other can be chained.
        """
        if state not in STATES:
            raise ValueError(
                f"{state} is not a valid state, use abbreviation like CA, ..."
            )
        validate_proportion_range(
            f"excluding state: {state}", min_proportion, max_proportion
        )

        self.excluded_geos.append(
            InputRange(
                name=state,
                min_proportion=min_proportion,
                max_proportion=max_proportion,
            )
        )
        return self

    def available_attributes(self):
        """
        Returns:
            List of all available attributes on which to filter.
        """
        return list(ATTRIBUTES.keys())

    def with_attributes(
        self,
        attribute: str,
        min: Union[float, int] = None,
        max: Union[float, int] = None,
    ):
        """
        Args:
            attribute: Pool attribute to filter on. For all possible attributes, run query.available_attributes()
            min_proportion: Minimum required for that attribute.
            max_proportion: Maximum required for that attribute.

        Returns:
            The query itself so these calls and other can be chained.
        """
        validate_range(f"attribute: {attribute}", min, max)
        if attribute not in ATTRIBUTES:
            raise ValueError(
                f"unknown attributes {attribute}, available are {','.join(ATTRIBUTES.keys())}"
            )
        self.attributes[attribute] = ATTRIBUTES[attribute](min, max)
        return self

    def available_predictions(self):
        """
        Returns:
            List of all available predictions on which to filter.
        """
        return list(PREDICTIONS.keys())

    def with_predictions(
        self,
        prediction: str,
        min: Union[float, int] = None,
        max: Union[float, int] = None,
    ):
        """
        Args:
            prediction: Pool prepayment prediction to filter on. For all possible attributes, run query.available_predictions()
            min_proportion: Minimum required for that prediction.
            max_proportion: Maximum required for that prediction.

        Returns:
            The query itself so these calls and other can be chained.
        """
        validate_range(f"prediction: {prediction}", min, max)
        if prediction not in PREDICTIONS:
            raise ValueError(
                f"unknown prediction {prediction}, available are {','.join(PREDICTIONS.keys())}"
            )
        self.predictions[prediction] = PREDICTIONS[prediction](min, max)
        return self

    def to_dict(self):
        out = {}
        if self.cohorts is not None:
            out["cohorts"] = {"cohorts": self.cohorts}
        if self.with_criteria():
            out["criteria"] = {}
        if self.with_servicers_criteria():
            servicers = {}
            if self.included_servicers:
                servicers["includes"] = [
                    inp.to_dict() for inp in self.included_servicers
                ]
            if self.excluded_servicers:
                servicers["excludes"] = [
                    inp.to_dict() for inp in self.excluded_servicers
                ]
            out["criteria"]["servicer"] = servicers
        if self.with_geos_criteria():
            geos = {}
            if self.included_geos:
                geos["includes"] = [inp.to_dict() for inp in self.included_geos]
            if self.excluded_geos:
                geos["excludes"] = [inp.to_dict() for inp in self.excluded_geos]
            out["criteria"]["geo"] = geos
        if self.with_attributes_criteria():
            out["attributes"] = {
                rg.name: rg.to_dict() for attribute, rg in self.attributes.items()
            }
        if self.with_predictions_criteria():
            out["predictions"] = {
                rg.name: rg.to_dict() for prediction, rg in self.predictions.items()
            }
        return out
